/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace SudokuSolver
{
    /// <summary>
    /// A small window that shows an animation while the program solves
    /// the current Sudoku. The progressbar shows how many threads are
    /// done with their workload. With the use of a timer the next animation
    /// picture is shown every 100 ms.
    /// </summary>
    public partial class SolveAnimationForm1 : Form
    {
        /// <summary>
        /// This Delegate is used by the showResults method to make it possible
        /// to invoke itself. This necessary so that the calculating thread can update
        /// the GUI.
        /// </summary>
        public delegate void SimpleMethod();

        /// <summary>
        /// This Delegate is used by the updateProgressBar method to make it possible
        /// to invoke itself. This necessary so that the calculating thread can update
        /// the progressbar in the animation window of the GUI.
        /// </summary>
        /// <param name="value">The percentage of the progress bar. This value should be between 0 and 100.</param>
        public delegate void SimpleIntMethod(int value);

        // holds the pictures for the animation and the current picture that is shown
        private Image[] m_imageList;
        private int m_currentImage = 0;
        private bool m_closeDown = false;

        /// <summary>
        /// Standard constructor initializes the GUI.
        /// </summary>
        public SolveAnimationForm1()
        {
            InitializeComponent();
        }

        /// <summary>
        /// This method gets called when the timer is activated and 100 ms have passed.
        /// It shows the next animation picture in the background of the window.
        /// </summary>
        private void timer1_Tick(object sender, EventArgs e)
        {
            // show next image in list
            m_currentImage++;
            if (m_currentImage >= 10) m_currentImage = 0;
            this.AnimationPanel1.BackgroundImage = m_imageList[m_currentImage];

            // close down window
            if (m_closeDown)
            {
                this.DialogResult = DialogResult.OK;
                this.Close();
            }
        }

        /// <summary>
        /// This method gets called when the animation window gets created. It loads the
        /// 10 animation pictures that are embedded as a resource from the executable.
        /// </summary>
        private void SolveAnimationForm1_Load(object sender, EventArgs e)
        {
            // load all animation images into one array
            m_imageList = new Image[10];
            System.IO.Stream st;
            System.Reflection.Assembly a = System.Reflection.Assembly.GetExecutingAssembly();
            for (int i = 0; i < 10; i++)
            {
                st = a.GetManifestResourceStream("SudukoSolver.Resources.SolvingAnimation" + Convert.ToString(i + 1) + ".png");
                m_imageList[i] = Image.FromStream(st);
            }

            // reset the state of the form
            reset();
        }

        /// <summary>
        /// This method gets called when the animation window gets closed. It deactivates
        /// the timer, so that no resources are wasted.
        /// </summary>
        private void SolveAnimationForm1_FormClosed(object sender, FormClosedEventArgs e)
        {
            // disable timer to stop animation
            timer1.Enabled = false;
        }

        /// <summary>
        /// This method enables the caller to update the progressbar. The value
        /// set as parameter sets the percentage of the progressbar.
        /// </summary>
        /// <param name="value">This integer value should be between 0 and 100.</param>
        public void setProgressBar(int value)
        {
            if (InvokeRequired) Invoke(new SimpleIntMethod(setProgressBar), new object[] { value });
            else
            {
                // updates the progressbar
                if (value < 0) value = 0;
                if (value > 100) value = 100;
                this.AnimationProgressBar1.Value = value;
            }
        }

        /// <summary>
        /// This method gets called from the solving thread to close down the animation window
        /// </summary>
        public void closeAnimationWindow()
        {
            if (InvokeRequired) Invoke(new SimpleMethod(closeAnimationWindow));
            else m_closeDown = true;
        }

        /// <summary>
        /// This method resets all animation and progressbar related values to their
        /// inital values. It also enables the timer.
        /// </summary>
        private void reset()
        {
            // reset to start conditions
            // show first image
            m_currentImage = 0;
            this.AnimationPanel1.BackgroundImage = m_imageList[m_currentImage];

            // reset the progressbar
            this.AnimationProgressBar1.Value = 0;

            // ativate timer for animation
            timer1.Enabled = true;

            // reset dialog result
            this.DialogResult = DialogResult.None;
        }

        /// <summary>
        /// This method gets called when the animation window is shown. It ensures,
        /// that the animations are reset before the window is shown.
        /// </summary>
        private void SolveAnimationForm1_Shown(object sender, EventArgs e)
        {
            // reset the state of the form
            reset();
        }

        /// <summary>
        /// This method gets called when the animation window is should be closed. This
        /// can happen through user input or by the end of the solving thread. It the user
        /// aborted the calculation, the parent window has to know to kill the solving threads.
        /// </summary>
        private void SolveAnimationForm1_FormClosing(object sender, FormClosingEventArgs e)
        {
            // if the user terminates the dialog, tell the parent window
            if (this.DialogResult == DialogResult.None) this.DialogResult = DialogResult.Cancel;
        }
    }
}