/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;

namespace SudokuSolver
{
    /// <summary>
    /// This class encapsulates a complete Sudoku board. It provides methods
    /// to fix numbers on the board as well as support methods for the solving
    /// algorithm.
    /// </summary>
    class SudokuCellField
    {
        /// <summary>
        /// 2-dimensional array to hold the 81 Sudoku cells.
        /// </summary>
        private SudokuCell[,] m_cells = new SudokuCell[9, 9];

        /// <summary>
        /// Standard constructor.
        /// </summary>
        public SudokuCellField()
        {
            for (int j = 0; j < 9; j++)
                for (int i = 0; i < 9; i++)
                    m_cells[j, i] = new SudokuCell();
        }

        /// <summary>
        /// Copy constructor.
        /// </summary>
        /// <param name="copy">The SudokuCellField that should be copied.</param>
        public SudokuCellField(SudokuCellField copy)
        {
            // use copy constructor for each cell
            for (int j = 0; j < 9; j++)
                for (int i = 0; i < 9; i++)
                    m_cells[j, i] = new SudokuCell(copy.m_cells[j, i]);
        }

        /// <summary>
        /// This method fixes one number on the indicated position of the Sudoku board.
        /// This will also exclude this number from all the cells on the same panel, in
        /// the same row and in the same column. Because of performance considerations, 
        /// the position and the number are not checked against boundaries.
        /// </summary>
        /// <param name="y">The y-coordinate of the Sudoku cell. Needs to be between 0 and 8.</param>
        /// <param name="x">The x-coordinate of the Sudoku cell. Needs to be between 0 and 8.</param>
        /// <param name="number">The number that should be fixed. Needs to be between 0 and 8.</param>
        public void fixNumber(int y, int x, int number)
        {
            // fixes the given number within the row, column and panel
            // because of performance considerations, the indices are not checked against boundaries
            // fix this number
            m_cells[y, x].fixNumber(number);

            // delete the number for the current panel
            int xmin = (x / 3) * 3;
            int xmax = xmin + 3;
            int ymin = (y / 3) * 3;
            int ymax = ymin + 3;
            for (int k = ymin; k < ymax; k++)
                for (int l = xmin; l < xmax; l++)
                    if (!((k == y) && (l == x))) m_cells[k, l].removePossibility(number);

            // delete the number for the current row
            for (int l = 0; l < 9; l++)
                if (l != x) m_cells[y, l].removePossibility(number);

            // delete the number for the current column
            for (int k = 0; k < 9; k++)
                if (k != y) m_cells[k, x].removePossibility(number);
        }

        /// <summary>
        /// This method checks if the indicated cell has a fixed number. If so, it returns
        /// the fixed number in the index parameter. Because of performance considerations, 
        /// the position is not checked against boundaries.
        /// </summary>
        /// <param name="y">The y-coordinate of the Sudoku cell. Needs to be between 0 and 8.</param>
        /// <param name="x">The x-coordinate of the Sudoku cell. Needs to be between 0 and 8.</param>
        /// <param name="index">The fixed number, if this cell has one.</param>
        /// <returns>True if the cell has a fixed number, False otherwise.</returns>
        public bool getNumber(int y, int x, ref int index)
        {
            if (!(m_cells[y, x].needsProcessing()))
            {
                index = m_cells[y, x].getFixedNumber();
                return true;
            }
            return false;
        }

        /// <summary>
        /// This method searches for cells that have only one possibility left.
        /// If it finds one, it removes the possibility of this number from the 
        /// cells in the current row, column and panel.
        /// </summary>
        /// <returns>The number cells that were fixed.</returns>
        public int findFixedPositions()
        {
            int foundPositions = 0;
            // run through all cells and check if they have a new fixed number
            for (int j = 0; j < 9; j++)
            {
                for (int i = 0; i < 9; i++)
                {
                    if (m_cells[j, i].checkForOnePossibilityLeft())
                    {
                        // found one, fix this number for column, row and panel
                        this.fixNumber(j, i, m_cells[j, i].getFixedNumber());
                        foundPositions++;
                    }
                }
            }

            // return the number of postions that were fixed
            return foundPositions;
        }

        /// <summary>
        /// This small struct is used to save the position and the possibilities
        /// of a pair that was found on the Sudoku board. A pair is a cell that
        /// has only two possibilities left.
        /// </summary>
        private struct pairInfo
        {
            public int x;
            public int y;
            public int pair;
        }

        /// <summary>
        /// This method checks the Sudoku board for naked pairs. Naked pairs are
        /// cells that are in the same row, column or panel that have only two
        /// possibilities left. If two pairs have the same possibilities within
        /// the same row, column or panel they exclude these possibilities for all
        /// the other cells in this row, column or panel.
        /// </summary>
        public void checkForNakedPairs()
        {
            // run through each column, row and panel and check
            // if there are naked pairs that remove posibilities in other cells

            // array that holds the information about the pairs
            pairInfo[] pairArray = new pairInfo[9];

            // check for each panel
            for (int p = 0; p < 3; p++)
            {
                for (int q = 0; q < 3; q++)
                {
                    int xmin = q * 3;
                    int xmax = xmin + 3;
                    int ymin = p * 3;
                    int ymax = ymin + 3;
                    int count = 0;

                    // count cells with pairs
                    for (int k = ymin; k < ymax; k++)
                    {
                        for (int l = xmin; l < xmax; l++)
                        {
                            // check if this cell has only 2 possibilities
                            if (m_cells[k, l].freeNumbers() == 2)
                            {
                                pairArray[count].y = k;
                                pairArray[count].x = l;
                                pairArray[count].pair = m_cells[k, l].possiblities();
                                count++;
                            }
                        }
                    }

                    // continue only, if there are at least 2 pairs
                    if (count >= 2)
                    {
                        // check all combination of pairs
                        for (int m = 0; m < (count - 1); m++)
                        {
                            for (int n = m + 1; n < count; n++)
                            {
                                if (pairArray[m].pair == pairArray[n].pair)
                                {
                                    // found a pair
                                    // what pair was found?
                                    int possibility1 = (pairArray[m].pair / 10) - 1;
                                    int possibility2 = (pairArray[m].pair % 10) - 1;

                                    // remove the possibilities of the pair from all other fields on the panel
                                    for (int k = ymin; k < ymax; k++)
                                    {
                                        for (int l = xmin; l < xmax; l++)
                                        {
                                            if ((k == pairArray[m].y) && (l == pairArray[m].x)) continue;
                                            if ((k == pairArray[n].y) && (l == pairArray[n].x)) continue;
                                            m_cells[k, l].removePossibility(possibility1);
                                            m_cells[k, l].removePossibility(possibility2);
                                        }
                                    }
                                    break;
                                }
                            }
                        }
                    }
                }

                // check for each row
                for (int j = 0; j < 9; j++)
                {
                    int count = 0;

                    // check this row
                    for (int i = 0; i < 9; i++)
                    {
                        // check if this cell has only 2 possibilities
                        if (m_cells[j, i].freeNumbers() == 2)
                        {
                            pairArray[count].x = i;
                            pairArray[count].y = j;
                            pairArray[count].pair = m_cells[j, i].possiblities();
                            count++;
                        }
                    }

                    // continue only, if there are at least 2 pairs
                    if (count >= 2)
                    {
                        // check all combination of pairs
                        for (int m = 0; m < (count - 1); m++)
                        {
                            for (int n = m + 1; n < count; n++)
                            {
                                if (pairArray[m].pair == pairArray[n].pair)
                                {
                                    // found a pair
                                    // what pair was found?
                                    int possibility1 = (pairArray[m].pair / 10) - 1;
                                    int possibility2 = (pairArray[m].pair % 10) - 1;

                                    // remove the possibilities of the pair from all other fields in the row
                                    for (int i = 0; i < 9; i++)
                                    {
                                        if ((j == pairArray[m].y) && (i == pairArray[m].x)) continue;
                                        if ((j == pairArray[n].y) && (i == pairArray[n].x)) continue;
                                        m_cells[j, i].removePossibility(possibility1);
                                        m_cells[j, i].removePossibility(possibility2);
                                    }
                                    break;
                                }
                            }
                        }
                    }
                }

                // check for each column
                for (int i = 0; i < 9; i++)
                {
                    int count = 0;

                    // check this row
                    for (int j = 0; j < 9; j++)
                    {
                        // check if this cell has only 2 possibilities
                        if (m_cells[j, i].freeNumbers() == 2)
                        {
                            pairArray[count].x = i;
                            pairArray[count].y = j;
                            pairArray[count].pair = m_cells[j, i].possiblities();
                            count++;
                        }
                    }

                    // continue only, if there are at least 2 pairs
                    if (count >= 2)
                    {
                        // check all combination of pairs
                        for (int m = 0; m < (count - 1); m++)
                        {
                            for (int n = m + 1; n < count; n++)
                            {
                                if (pairArray[m].pair == pairArray[n].pair)
                                {
                                    // found a pair
                                    // what pair was found?
                                    int possibility1 = (pairArray[m].pair / 10) - 1;
                                    int possibility2 = (pairArray[m].pair % 10) - 1;

                                    // remove the possibilities of the pair from all other fields in the row
                                    for (int j = 0; j < 9; j++)
                                    {
                                        if ((j == pairArray[m].y) && (i == pairArray[m].x)) continue;
                                        if ((j == pairArray[n].y) && (i == pairArray[n].x)) continue;
                                        m_cells[j, i].removePossibility(possibility1);
                                        m_cells[j, i].removePossibility(possibility2);
                                    }
                                    break;
                                }
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// This method checks for number and for each row, column and panel if there
        /// is only one cell left where the number is possible. If it finds such a 
        /// place, it fixes the number for this cell.
        /// </summary>
        /// <returns>The number cells that were fixed.</returns>
        public int checkNumberPositions()
        {
            int foundPositions = 0;
            // run through all numbers and check each column, row and panel
            // if there is only one position left, where a certain number can be

            // check for each number
            for (int n = 0; n < 9; n++)
            {
                // check for each panel, if there is only one field left
                for (int p = 0; p < 3; p++)
                {
                    for (int q = 0; q < 3; q++)
                    {
                        int xmin = q * 3;
                        int xmax = xmin + 3;
                        int ymin = p * 3;
                        int ymax = ymin + 3;
                        int count = 0;
                        int posx = 0;
                        int posy = 0;

                        // check the panel
                        for (int k = ymin; k < ymax; k++)
                        {
                            for (int l = xmin; l < xmax; l++)
                            {
                                // check if this number is possible in this cell
                                if (m_cells[k, l].isPossible(n))
                                {
                                    count++;
                                    posx = l;
                                    posy = k;
                                }
                            }
                        }

                        // if there is only one place left, check if it is already fixed
                        if ((count == 1) && (m_cells[posy, posx].needsProcessing()))
                        {
                            // otherwise fix this cell
                            fixNumber(posy, posx, n);
                            foundPositions++;
                        }
                    }
                }

                // check for each row, if there is only one field left
                for (int j = 0; j < 9; j++)
                {
                    int count = 0;
                    int posx = 0;
                    int posy = 0;

                    // check this row
                    for (int i = 0; i < 9; i++)
                    {
                        if (m_cells[j, i].isPossible(n))
                        {
                            count++;
                            posx = i;
                            posy = j;
                        }
                    }

                    // if there is only one place left, check if it is already fixed
                    if ((count == 1) && (m_cells[posy, posx].needsProcessing()))
                    {
                        // otherwise fix this cell
                        fixNumber(posy, posx, n);
                        foundPositions++;
                    }
                }

                // check for each column, if there is only one field left
                for (int i = 0; i < 9; i++)
                {
                    int count = 0;
                    int posx = 0;
                    int posy = 0;

                    // check this column
                    for (int j = 0; j < 9; j++)
                    {
                        if (m_cells[j, i].isPossible(n))
                        {
                            count++;
                            posx = i;
                            posy = j;
                        }
                    }

                    // if there is only one place left, check if it is already fixed
                    if ((count == 1) && (m_cells[posy, posx].needsProcessing()))
                    {
                        fixNumber(posy, posx, n);
                        foundPositions++;
                    }
                }
            }

            // return the number of postions that were fixed
            return foundPositions;
        }

        /// <summary>
        /// This method checks if there there are still cells that need further
        /// processing.
        /// </summary>
        /// <returns>True if this boards needs further processing, False otherwise.</returns>
        public bool stillOpenCells()
        {
            // checks if there are still cells left, that have not been fixed
            bool foundOne = false;
            for (int j = 0; j < 9; j++)
                for (int i = 0; i < 9; i++)
                    if (m_cells[j, i].needsProcessing()) foundOne = true;

            return foundOne;
        }

        /// <summary>
        /// This method checks if the indicated cell needs further processing. Because of
        /// performance considerations, the position is not checked against boundaries.
        /// </summary>
        /// <param name="y">The y-coordinate of the Sudoku cell. Needs to be between 0 and 8.</param>
        /// <param name="x">The x-coordinate of the Sudoku cell. Needs to be between 0 and 8.</param>
        /// <returns>True if the indicated cells needs further processing, False otherwise.</returns>
        public bool needsProcessing(int y, int x)
        {
            return m_cells[y, x].needsProcessing();
        }

        /// <summary>
        /// This method checks if for the indicated cell the supplied number is still possible.
        /// Because of performance considerations, the position as well as the number are not 
        /// checked against boundaries.
        /// </summary>
        /// <param name="y">The y-coordinate of the Sudoku cell. Needs to be between 0 and 8.</param>
        /// <param name="x">The x-coordinate of the Sudoku cell. Needs to be between 0 and 8.</param>
        /// <param name="number">The number that should be checked as a possibility. Needs to be between 0 and 8.</param>
        /// <returns>True if the indicated number is possible for the indicated cell, False otherwise.</returns>
        public bool isPossible(int y, int x, int number)
        {
            return m_cells[y, x].isPossible(number);
        }

        /// <summary>
        /// This method checks if all cells of the Sudoku board are valid.
        /// </summary>
        /// <returns>True if the board is valid, False otherwise.</returns>
        public bool isValid()
        {
            // check each cell if it is valid
            for (int j = 0; j < 9; j++)
                for (int i = 0; i < 9; i++)
                    if (!(m_cells[j, i].isValid())) return false;
            return true;
        }

        /// <summary>
        /// This overloaded operator is implemented so that two fields can be compared.
        /// </summary>
        /// <param name="field1">First field to compare.</param>
        /// <param name="field2">Second field to compare</param>
        /// <returns>True if the fields are equal, False otherwise.</returns>
        public static bool operator ==(SudokuCellField field1, SudokuCellField field2) {
            for (int j = 0; j < 9; j++)
                for (int i = 0; i < 9; i++)
                    if (field1.m_cells[j, i] != field2.m_cells[j, i]) return false;
            return true;
        }

        /// <summary>
        /// This overloaded operator needs to be implemented as soon as the
        /// ==-operator is implemented.
        /// </summary>
        /// <param name="field1">First field to compare.</param>
        /// <param name="field2">Second field to compare</param>
        /// <returns>True if the fields are not equal, False otherwise.</returns>
        public static bool operator !=(SudokuCellField field1, SudokuCellField field2)
        {
            return !(field1 == field2);
        }

        /// <summary>
        /// This overloaded operator needs to be implemented as soon as the 
        /// ==-operator is implemented. It returns, if two objects are equal.
        /// </summary>
        /// <param name="obj">The other object that will be compared to this field.</param>
        /// <returns>True if the two objects are equal, False otherwise.</returns>
        public override bool Equals(object obj)
        {
            if (obj == null) return false;

            if (this.GetType() != obj.GetType()) return false;

            // safe because of the GetType check
            SudokuCellField field = (SudokuCellField)obj;

            // check members
            for (int j = 0; j < 9; j++) 
                for(int i = 0; i < 9; i++) 
                    if (!Object.Equals(m_cells[j,i], field.m_cells[j,i])) return false;

            return true;
        }

        /// <summary>
        /// This Method needs to be overriden to resolve a warning. It is not implemented
        /// and will raise a NotImplementedException if called anyway. This method will
        /// only be called if the Sudoku field class is used in a hashmap.
        /// </summary>
        /// <returns>Nothing, throws Exception if called.</returns>
        public override int GetHashCode()
        {
            throw new NotImplementedException();
        }
    }
}
